"""Structures used for defining queries."""

import dataclasses as dc
from typing import Any, ClassVar, Dict, Tuple

import pydantic.dataclasses as pdc
from pydantic import ConfigDict
from typing_extensions import dataclass_transform

QUERY_ATTRIB = "__query__"


class _Calculated:
    """Marker class for calculated values."""


# Typehint the class marker to Any to avoid conflict with the type checkers
Calculated: Any = _Calculated


@dataclass_transform()
class QueryMeta(type):
    """Metaclass for generating Query objects that the library supports.

    Classes using this metaclass are automatically converted to Pydantic
    Dataclasses for validation support. Also, the `__query__` attribute is
    verified to be provided either as a value or as a type hint.

    Raises:
        ValueError: When the `__query__` attribute is not defined.

    Returns:
        _type_: A new type compatible for the library query functionalities.
    """

    @staticmethod
    def _check_query_attribute(dataclass: "QueryMeta", dct: Dict[str, Any]):
        try:
            if isinstance(getattr(dataclass, QUERY_ATTRIB), str):
                return True
        except AttributeError:
            return QUERY_ATTRIB in dct["__annotations__"]

        return False

    @staticmethod
    def _resolve_calculated_fields(dct: Dict[str, Any]):
        for k, v in dct.items():
            if v == Calculated:
                dct[k] = dc.field(init=False)

    def __new__(cls, name: str, bases: Tuple[type, ...], dct: Dict[str, Any]):
        """Creates a new class of the metaclass.

        This wraps the newly created class with Pydantic Dataclass and checks
        the `__query__` attribute.
        """
        cls._resolve_calculated_fields(dct)
        dataclass = super().__new__(cls, name, bases, dct)
        if not cls._check_query_attribute(dataclass, dct):
            raise ValueError(f"A string {QUERY_ATTRIB} must be provided")
        return pdc.dataclass(dataclass, config=ConfigDict(arbitrary_types_allowed=True))


class Query(metaclass=QueryMeta):
    """Base class for query definitions.

    This class can be inherited instead of providing the `QueryMeta` as the
    class metaclass.
    """

    __query__: ClassVar[str]



"""Main entry point for the altqq library."""

from altqq.structs import Calculated, Query
from altqq.translators.mysql import MySQLQuery, MySQLTranslator
from altqq.translators.plain_text import PlainTextTranslator
from altqq.translators.psycopg import PsycopgQuery, PsycopgTranslator
from altqq.translators.pyodbc import PyODBCQuery, PyODBCTranslator
from altqq.types import ListParameter, NonParameter

__all__ = [
    "Query",
    "Calculated",
    "NonParameter",
    "ListParameter",
    "to_pyodbc",
    "to_psycopg",
    "to_mysql",
    "to_plain_text",
]


class Translators:
    """Definition of available translators."""

    PYODBC = PyODBCTranslator()
    PSYCOPG = PsycopgTranslator()
    MYSQL = MySQLTranslator()
    PLAIN_TEXT = PlainTextTranslator()


def to_pyodbc(query: Query) -> PyODBCQuery:
    """Converts a `Query` to its corresponding `PyODBCQuery` object.

    Args:
        query (Query): Query to translate to PyODBC

    Returns:
        PyODBCQuery: Equivalent query for PyODBC usage.
    """
    return Translators.PYODBC(query)


def to_psycopg(query: Query) -> PsycopgQuery:
    """Converts a `Query` to its corresponding `PsycopgQuery` object.

    Args:
        query (Query): Query to translate to Psycopg

    Returns:
        PsycopgQuery: Equivalent query for Psycopg usage.
    """
    return Translators.PSYCOPG(query)


def to_mysql(query: Query) -> MySQLQuery:
    """Converts a `Query` to its corresponding `MySQL` object.

    Args:
        query (Query): Query to translate to MySQL

    Returns:
        MySQLQuery: Equivalent query for MySQL usage.
    """
    return Translators.MYSQL(query)


def to_plain_text(query: Query) -> str:
    """Converts a `Query` to a plain text SQL.

    The conversion to plain text also handles some of the data types. None
    is converted to `NULL`, numeric values are written as they are and
    string values and other object types are escaped using `'`.

    Args:
        query (Query): Query to convert.

    Returns:
        str: Query as plain text.
    """
    return Translators.PLAIN_TEXT(query)



"""Module for converting Query objects for PyODBC execution."""

import dataclasses as dc
from string import Formatter
from typing import Any, Iterable, List, Mapping, Sequence, Union

from altqq.structs import Query
from altqq.translators import common
from altqq.types import QueryValueTypes, T


@dc.dataclass
class PyODBCStatement:
    """Represents a generated PyODBC statement."""

    statement: Any
    parameters: Iterable[Any]


@dc.dataclass
class PyODBCQuery:
    """Converted `Query` object for PyODBC usage."""

    query: str
    parameters: Iterable[Any]


class PyODBCFormatter(Formatter):
    """Converts `PyODBCStatement` objects to query string.

    Aside from converting the `PyODBCStatement`, this class also tracks the
    parameter substitutions during the formatting. These values are stored
    in the `parameter_values` attribute.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        """Resets the state of the formatter."""
        self.parameter_values: List[Any] = []

    def get_value(
        self,
        key: Union[str, int],
        args: Sequence[Any],
        kwargs: Mapping[str, PyODBCStatement],
    ) -> Any:
        """Retrieves the value formatting values from the statement mapping.

        Args:
            key (Union[str, int]): Key found in the raw string. Always `str`.
            args (Sequence[Any]): Ordered arguments. Unused.
            kwargs (Mapping[str, PyODBCStatement]): Key mapping to the
                PyODBCStatements.

        Returns:
            Any: The statement value provided in the PyODBCStatement.
        """
        # The args section is never used
        assert isinstance(key, str)

        field = kwargs[key]
        self.parameter_values.extend(field.parameters)
        return field.statement


class PyODBCTranslator:
    """Converts a `Query` to its corresponding `PyODBCQuery` object."""

    MARKER = "?"

    def _resolve_value(self, query: Query, field: dc.Field[T]) -> PyODBCStatement:
        value = getattr(query, field.name)
        if common.is_query_instance(value):
            qq = self.__call__(value)
            return PyODBCStatement(qq.query, qq.parameters)

        # Field has no typing in older python versions
        field_type = common.get_parameter_type(field.type)  # type: ignore
        if field_type == QueryValueTypes.NON_PARAMETER:
            return PyODBCStatement(value, ())
        elif field_type == QueryValueTypes.LIST_PARAMETER:
            return PyODBCStatement(
                common.create_list_markers(self.MARKER, len(value)), value
            )
        else:
            return PyODBCStatement(self.MARKER, [value])

    def __call__(self, query: Query) -> PyODBCQuery:
        """Converts a `Query` to its corresponding `PyODBCQuery` object.

        Args:
        query (Query): Query to translate to PyODBC

        Returns:
            PyODBCQuery: Equivalent query for PyODBC usage.
        """
        formatter = PyODBCFormatter()
        fields = dc.fields(query)
        format_dict = {f.name: self._resolve_value(query, f) for f in fields}

        return PyODBCQuery(
            query=formatter.format(query.__query__, **format_dict),
            parameters=formatter.parameter_values,
        )



"""Translator related functions not belonging to other areas."""

import typing
from typing import Any, Set, Type, cast

from typing_extensions import Annotated

from altqq.structs import Query
from altqq.types import QueryValueTypes


def is_query_instance(value: Any) -> bool:
    """Checks if the value is a subclass of altqq.structs.Query.

    Args:
        value (Any): Object to check.

    Returns:
        bool: True if a subclass, else False.
    """
    return isinstance(value, Query)


ANNOTATED_TYPES: Set[QueryValueTypes] = {
    QueryValueTypes.NON_PARAMETER,
    QueryValueTypes.LIST_PARAMETER,
}


def get_parameter_type(cls: Type[Any]) -> QueryValueTypes:
    """Extracts the value type based on the typing.

    Args:
        cls (Type[Any]): Type to check.

    Returns:
        QueryValueTypes: Role of the value in the query
    """
    if typing.get_origin(cls) == Annotated:
        # Pyright can't match the __metadata__ attribute to Annotated
        for metadata in cls.__metadata__:  # type: ignore
            if metadata in ANNOTATED_TYPES:
                return cast(QueryValueTypes, metadata)

    return QueryValueTypes.PARAMETER


def create_list_markers(marker: str, n: int) -> str:
    """Creates list markers for list parameter types.

    Args:
        marker (str): Parameter marker used by the translation
        n (int): Number of parameters in the list

    Returns:
        str: String to represent the list
    """
    comma_separated = ",".join(marker for _ in range(n))
    return f"({comma_separated})"



"""Module for converting Query objects to plain text SQL."""

import dataclasses as dc
from typing import Any

from altqq.structs import Query
from altqq.translators import common
from altqq.types import QueryValueTypes, T


class PlainTextTranslator:
    """Converts a `Query` to a plain text SQL."""

    def _resolve_parameters(self, value: Any) -> str:
        # Numeric types are not escaped
        if isinstance(value, (int, float)):
            return str(value)

        # None is written as NULL
        if value is None:
            return "NULL"

        # All other types fall down to strings and are escaped
        return f"'{value}'"

    def _resolve_value(self, query: Query, field: dc.Field[T]) -> str:
        value = getattr(query, field.name)
        if common.is_query_instance(value):
            return self.__call__(value)

        # Field has no typing in older python versions
        field_type = common.get_parameter_type(field.type)  # type: ignore
        if field_type == QueryValueTypes.NON_PARAMETER:
            return value
        elif field_type == QueryValueTypes.LIST_PARAMETER:
            comma_separated = ",".join(self._resolve_parameters(p) for p in value)
            return f"({comma_separated})"
        else:
            return self._resolve_parameters(value)

    def __call__(self, query: Query) -> str:
        """Converts a `Query` to a plain text SQL.

        The conversion to plain text also handles some of the data types. None
        is converted to `NULL`, numeric values are written as they are and
        string values and other object types are escaped using `'`.

        Args:
            query (Query): Query to convert.

        Returns:
            str: Query as plain text.
        """
        fields = dc.fields(query)
        format_dict = {f.name: self._resolve_value(query, f) for f in fields}
        return query.__query__.format(**format_dict)



"""Module for converting Query objects for mysql execution."""

# For now the psycopg classes are being reused for mysql since they use
# similar formatting. This can changed if there are edge cases that were
# missed so separate classes will be introduced for it

import dataclasses as dc
from typing import Any, Tuple

from altqq.structs import Query
from altqq.translators.psycopg import PsycopgTranslator


@dc.dataclass
class MySQLQuery:
    """Converted `Query` object for MySQL usage."""

    query: str
    parameters: Tuple[Any, ...]


class MySQLTranslator:
    """Converts a `Query` to its corresponding `MySQLQuery` object."""

    def __init__(self):
        self._translator = PsycopgTranslator()

    def __call__(self, query: Query) -> MySQLQuery:
        """Converts a `Query` to its corresponding `MySQL` object.

        Args:
            query (Query): Query to translate to MySQL

        Returns:
            MySQLQuery: Equivalent query for MySQL usage.
        """
        psql_query = self._translator(query)
        return MySQLQuery(query=psql_query.query, parameters=psql_query.parameters)
